from typing import Tuple, List, Optional

from .ILayerListener import ILayerListener
from ..Listenable import Listenable
from ..constants import Direction
from ..constants.CellValue import CellValue


class Layer(Listenable[ILayerListener]):
    def __init__(self, size: Tuple[int, int], defaultValue: CellValue):
        super().__init__()
        self.__size = size
        self.__defaultValue = defaultValue
        self.__cells = []
        for y in range(size[1]):
            row = [defaultValue] * size[0]
            self.__cells.append(row)

    @property
    def size(self) -> Tuple[int, int]:
        return self.__size

    @property
    def width(self) -> int:
        return self.__size[0]

    @property
    def height(self) -> int:
        return self.__size[1]

    def __getitem__(self, coords: Tuple[int, int]) -> CellValue:
        return self.getValue(coords)

    def getValue(self, coords: Tuple[int, int], direction: Optional[Direction] = None) -> CellValue:
        x, y = coords[0], coords[1]
        w, h = self.__size
        assert 0 <= x < w, f"Invalid x={x}"
        assert 0 <= y < h, f"Invalid y={y}"
        if direction:
            if direction == Direction.LEFT:
                if x < 1:
                    return self.__defaultValue
                return self.__cells[y][x - 1]
            if direction == Direction.RIGHT:
                if x >= w - 1:
                    return self.__defaultValue
                return self.__cells[y][x + 1]
            if direction == Direction.TOP:
                if y < 1:
                    return self.__defaultValue
                return self.__cells[y - 1][x]
            if direction == Direction.BOTTOM:
                if y >= h - 1:
                    return self.__defaultValue
                return self.__cells[y + 1][x]
        return self.__cells[y][x]

    def __setitem__(self, coords: Tuple[int, int], value: CellValue):
        self.setValue(coords, value)

    def setValue(self, coords: Tuple[int, int], value: CellValue):
        x, y = coords[0], coords[1]
        assert 0 <= x < self.__size[0], f"Invalid x={x}"
        assert 0 <= y < self.__size[1], f"Invalid y={y}"
        self.__cells[y][x]= value

    def fill(self, value: CellValue):
        for y in range(0, self.height):
            for x in range(0, self.width):
                self.__cells[y][x] = value

    # Codes

    def getNeighbors4(self, cell: Tuple[int, int]) -> Tuple[int, int, int, int]:
        x, y = cell
        w, h = self.__size[0] - 1, self.__size[1] - 1
        left = self.__cells[y][x - 1] if x > 0 else self.__defaultValue
        right = self.__cells[y][x + 1] if x < w else self.__defaultValue
        top = self.__cells[y - 1][x] if y > 0 else self.__defaultValue
        bottom = self.__cells[y + 1][x] if y < h else self.__defaultValue
        return left, top, bottom, right

    def getNeighbors8(self, cell: Tuple[int, int]) -> Tuple[int, int, int, int, int, int, int, int]:
        x, y = cell
        w, h = self.__size[0] - 1, self.__size[1] - 1
        topLeft = self.__cells[y - 1][x - 1] if y > 0 and x > 0 else self.__defaultValue
        top = self.__cells[y - 1][x] if y > 0 else self.__defaultValue
        topRight = self.__cells[y - 1][x + 1] if y > 0 and x < w else self.__defaultValue
        left = self.__cells[y][x - 1] if x > 0 else self.__defaultValue
        right = self.__cells[y][x + 1] if x < w else self.__defaultValue
        bottomLeft = self.__cells[y + 1][x - 1] if y < h and x > 0 else self.__defaultValue
        bottom = self.__cells[y + 1][x] if y < h else self.__defaultValue
        bottomRight = self.__cells[y + 1][x + 1] if y < h and x < w else self.__defaultValue
        return (topLeft, left, bottomLeft, top,
                bottom, topRight, right, bottomRight)

    # Layer listener

    def notifyCellChanged(self, cell: Tuple[int, int]):
        for listener in self.listeners:
            listener.cellChanged(self, cell)
