from typing import Tuple, Union

import pygame

from core.constants import CellValue
from core.logic import Logic
from core.state import World
from .GameMode import GameMode
from ..Mouse import Mouse
from ..component.IComponentListener import IComponentListener
from ..component.frame.PaletteFrame import PaletteFrame
from ..component.world.WorldComponent import WorldComponent
from ..theme.Theme import Theme


class EditGameMode(GameMode, IComponentListener):

    def __init__(self, theme: Theme, world: World):
        super().__init__(theme)
        self.__world = world
        self.__logic = Logic(world)
        self.__font = theme.getFont("default")

        # Components
        self.__worldComponent = WorldComponent(theme, world)

        self.__paletteFrame = PaletteFrame(theme, world)
        self.__paletteFrame.moveRelativeTo("bottom", self, "bottom")

        # Components order
        self.addComponent(self.__worldComponent)
        self.addComponent(self.__paletteFrame)

        # Brush
        self.__mainBrushLayer = "ground"
        self.__mainBrushValue: Union[int, str] = CellValue.GROUND_EARTH
        self.__secondaryBrushLayer = "ground"
        self.__secondaryBrushValue: Union[int, str] = CellValue.GROUND_SEA

        # Listener
        self.__worldComponent.registerListener(self)
        self.__paletteFrame.registerListener(self)

    # Game loop

    def dispose(self):
        self.__worldComponent.removeListener(self)
        self.__paletteFrame.removeListener(self)
        super().dispose()

    def update(self):
        self.__logic.executeCommands()

    # UI Event handler

    def keyDown(self, key: int):
        if key == pygame.K_F1:
            self.__worldComponent.setAutoTiling(True)
            return True
        elif key == pygame.K_F2:
            self.__worldComponent.setAutoTiling(False)
            return True
        else:
            return super().keyDown(key)

    # Component listener

    def __updateCell(self, cell: Tuple[int, int], mouse: Mouse):
        if mouse.button1 or mouse.button2:
            brushLayer = self.__mainBrushLayer
            brushValue = self.__mainBrushValue
            fill = mouse.button2
        elif mouse.button3:
            brushLayer = self.__secondaryBrushLayer
            brushValue = self.__secondaryBrushValue
            fill = False
        else:
            return
        Command = self.__logic.getSetLayerValueCommand(brushLayer)
        command = Command(cell, brushValue, fill)
        self.__logic.addCommand(command)

    def worldCellClicked(self, cell: Tuple[int, int], mouse: Mouse):
        self.__updateCell(cell, mouse)

    def worldCellEntered(self, cell: Tuple[int, int], mouse: Mouse, dragging: bool):
        if dragging:
            self.__updateCell(cell, mouse)

    def mainBrushSelected(self, layerName: str, value: Union[int, str]):
        self.__mainBrushLayer = layerName
        self.__mainBrushValue = value

    def secondaryBrushSelected(self, layerName: str, value: Union[int, str]):
        self.__secondaryBrushLayer = layerName
        self.__secondaryBrushValue = value
