import os
from typing import Dict, Tuple, Union, Optional

import pygame
from pygame.font import Font
from pygame.surface import Surface

from core.constants import CellValue
from .Tileset import Tileset


class Theme:
    def __init__(self):
        self.__viewSize = (1024, 768)

        # Tilesets
        self.__surfaces: Dict[str, Surface] = {}
        self.__tilesets: Dict[str, Tileset] = {}
        for name, tilesDef in tilesDefs.items():
            tileset = Tileset(self, tilesDef["tileSize"], tilesDef["imageFile"])
            tileset.addTiles(tilesDef["tiles"])
            self.__tilesets[name] = tileset

        self.__frameBorderSize = self.__tilesets["frame"].tileSize[0]

        # Text
        self.__fontsDef = {
            "default": {
                "file": "font/prstart/prstartk.ttf",
                "size": 8
            }
        }
        self.__fontColors = {
            "default": (100, 75, 50)
        }
        self.__fonts = {}

    @property
    def frameBorderSize(self) -> int:
        return self.__frameBorderSize

    @property
    def viewSize(self) -> Tuple[int, int]:
        return self.__viewSize

    @viewSize.setter
    def viewSize(self, size: Tuple[int, int]):
        self.__viewSize = size

    def getSurface(self, imageFile: str) -> Surface:
        if imageFile not in self.__surfaces:
            fullPath = os.path.join("assets", imageFile)
            if not os.path.exists(fullPath):
                raise ValueError(f"No file '{fullPath}'")
            self.__surfaces[imageFile] = pygame.image.load(fullPath).convert_alpha()
        return self.__surfaces[imageFile]

    def getTileset(self, name: str) -> Tileset:
        if name not in self.__tilesets:
            raise ValueError(f"No tileset {name}")
        return self.__tilesets[name]

    def getFont(self, name: Optional[str] = None) -> Font:
        if name is None:
            name = "default"
        if name not in self.__fontsDef:
            raise ValueError("No font {}".format(name))
        fontDef = self.__fontsDef[name]
        file = os.path.join("assets", fontDef["file"])
        size = fontDef["size"]
        fontId = file, size
        if fontId not in self.__fonts:
            self.__fonts[fontId] = pygame.font.Font(file, size)
        return self.__fonts[fontId]

    def getFontColor(self, name: Union[None, str] = None) -> Tuple[int, int, int]:
        if name is None:
            name = "default"
        if name not in self.__fontColors:
            raise ValueError("No font color {}".format(name))
        return self.__fontColors[name]


tilesDefs = {
    "ground": {
        "imageFile": "toen/ground.png",
        "tileSize": (16, 16),
        "tiles": {
            CellValue.GROUND_SEA: [(4, 7), (5, 7), (6, 7), (7, 7)],
            CellValue.GROUND_EARTH: [(0, 7), (1, 7), (2, 7), (3, 7)],
        }
    },
    "impassable": {
        "imageFile": "toen/impassable.png",
        "tileSize": (16, 16),
        "tiles": {
            CellValue.NONE: (0, 0),
            CellValue.IMPASSABLE_RIVER: (0, 1),
            CellValue.IMPASSABLE_POND: [(1, 0), (2, 0), (3, 0)],
            CellValue.IMPASSABLE_MOUNTAIN: [(4, 0), (5, 0), (6, 0)],
            "riverMouthRight": (4, 1),
            "riverMouthLeft": (4, 2),
            "riverMouthTop": (4, 3),
            "riverMouthBottom": (4, 4),
        }
    },
    "objects": {
        "imageFile": "toen/objects.png",
        "tileSize": (16, 16),
        "tiles": {
            CellValue.NONE: (0, 0),
            CellValue.OBJECTS_SIGN: (1, 0),
            CellValue.OBJECTS_HILL: [(4, 0), (5, 1)],
            CellValue.OBJECTS_ROCKS: [(6, 1), (7, 1)],
            CellValue.OBJECTS_TREES: [(5, 0), (6, 0), (7, 0)],
            CellValue.OBJECTS_MILL: [(0, 1), (1, 1), (2, 1), (3, 1)],
            CellValue.OBJECTS_HOUSES: [(1, 2), (2, 2), (3, 2), (4, 2), (5, 2), (6, 2), (7, 2)],
            CellValue.OBJECTS_ROAD_DIRT: (0, 3),
            CellValue.OBJECTS_ROAD_STONE: (4, 3),
            CellValue.OBJECTS_FARM: [(8, 0), (9, 0), (10, 0), (11, 0)],
            CellValue.OBJECTS_CAMP: (8, 1),
            "bridgeDirt": [(0, 7), (1, 7)],
            "bridgeStone": [(4, 7), (5, 7)],
        },
    },
    "frame": {
        "imageFile": "toen/frame.png",
        "tileSize": (4, 4),
        "tiles": {
            "none": (0, 0),
            "topLeft": (0, 1),
            "top": (1, 1),
            "topRight": (2, 1),
            "left": (0, 2),
            "center": (1, 2),
            "right": (2, 2),
            "bottomLeft": (0, 3),
            "bottom": (1, 3),
            "bottomRight": (2, 3),
        },
    },
}

